import { toNumber } from 'lodash'
import { markRaw } from 'vue'

import IntegerField from './IntegerField'
import FloatField from './FloatField'
import ChoicesField from './ChoicesField'
import StringField from './StringField'
import BooleanField from './BooleanField'
import DictField from './DictField'
import ListField from './ListField'

export default {
  int: {
    // Mark the component as an object that should not be made reactive by Vue, to remove a warning about possible performance issues
    component: markRaw(IntegerField),
    validate (value) {
      const parsed = toNumber(value)
      if (!Number.isInteger(parsed)) throw new Error('Value must be a valid integer.')
      return parsed
    }
  },
  float: {
    component: markRaw(FloatField),
    validate (value) {
      const parsed = toNumber(value)
      if (!Number.isFinite(parsed)) throw new Error('Value must be a valid float.')
      return parsed
    }
  },
  string: {
    component: markRaw(StringField),
    validate (value) {
      return value
    }
  },
  enum: {
    component: markRaw(ChoicesField),
    validate (value, field) {
      if (!(field.choices.includes(value))) throw new Error(`${value} is not a valid option.`)
      return value
    }
  },
  bool: {
    component: markRaw(BooleanField),
    validate (value) {
      if (typeof value !== 'boolean') throw new Error('Value must be a valid boolean.')
      return value
    }
  },
  dict: {
    component: markRaw(DictField),
    validate (value) {
      if (typeof value !== 'object' | Object.getPrototypeOf(value) !== Object.prototype) throw new Error('Value must be a valid dictionary.')
      // Values should be of type String
      return Object.fromEntries(Object.entries(value).map(([k, v]) => [k, String(v).toString()]))
    }
  },
  list: {
    component: markRaw(ListField),
    validate (value, field) {
      if (!Array.isArray(value) || Object.values(value).some(value => value instanceof Error)) throw new Error(`Value must be a valid list of ${field?.subtype}.`)
      return value
    }
  }
}
