import { assert } from 'chai'
import axios from 'axios'
import { initialState, mutations } from '@/store/corpora.js'
import store from './index.spec.js'
import { jobsSample, exportSample } from '../samples.js'
import { assertRejects, FakeAxios } from '../testhelpers.js'
import { createPinia, setActivePinia } from 'pinia'
import { useJobsStore } from '@/stores'

describe('corpora', () => {
  describe('mutations', () => {
    it('set', () => {
      const state = {
        corpora: {
          corpusid: { id: 'corpusid' }
        },
        corporaLoaded: false
      }
      const corporaList = [
        {
          id: 'corpus1',
          name: 'Corpus 1',
          rights: ['read', 'write'],
          types: [
            {
              slug: 'type1',
              display_name: 'Type 1'
            },
            {
              slug: 'type2',
              display_name: 'Type 2'
            }
          ]
        }
      ]
      const expected = {
        corpora: {
          corpus1: {
            id: 'corpus1',
            name: 'Corpus 1',
            rights: ['read', 'write'],
            types: {
              type1: {
                slug: 'type1',
                display_name: 'Type 1'
              },
              type2: {
                slug: 'type2',
                display_name: 'Type 2'
              }
            }
          }
        },
        corporaLoaded: true
      }

      mutations.set(state, corporaList)
      assert.deepStrictEqual(state, expected)
    })

    describe('update', () => {
      it('updates an existing corpus', () => {
        const state = {
          corpora: {
            corpus1: {
              id: 'corpus1',
              name: 'Corpus 1'
            }
          }
        }

        mutations.update(state, {
          id: 'corpus1',
          name: 'New name',
          types: [
            {
              slug: 'type1',
              display_name: 'Type 1'
            },
            {
              slug: 'type2',
              display_name: 'Type 2'
            }
          ]
        })

        assert.deepStrictEqual(state, {
          corpora: {
            corpus1: {
              id: 'corpus1',
              name: 'New name',
              types: {
                type1: {
                  slug: 'type1',
                  display_name: 'Type 1'
                },
                type2: {
                  slug: 'type2',
                  display_name: 'Type 2'
                }
              }
            }
          }
        })
      })

      it('adds an unknown corpus', () => {
        const state = {
          corpora: {
            corpus1: {
              id: 'corpus1',
              name: 'Corpus 1'
            }
          }
        }

        mutations.update(state, {
          id: 'corpus2',
          name: 'New name'
        })

        assert.deepStrictEqual(state, {
          corpora: {
            corpus1: {
              id: 'corpus1',
              name: 'Corpus 1'
            },
            corpus2: {
              id: 'corpus2',
              name: 'New name'
            }
          }
        })
      })
    })

    describe('updateType', () => {
      it('updates an existing type on a corpus', () => {
        const state = {
          corpora: {
            corpus1: {
              id: 'corpus1',
              types: {
                type1: {
                  id: 'typeid',
                  slug: 'type1',
                  display_name: 'Type 1'
                }
              }
            }
          }
        }

        mutations.updateType(state, {
          corpus: 'corpus1',
          id: 'typeid',
          slug: 'type2',
          display_name: 'Type 2'
        })

        assert.deepStrictEqual(state, {
          corpora: {
            corpus1: {
              id: 'corpus1',
              types: {
                type2: {
                  id: 'typeid',
                  slug: 'type2',
                  display_name: 'Type 2'
                }
              }
            }
          }
        })
      })

      it('creates a new type on an existing corpus', () => {
        const state = {
          corpora: {
            corpus1: {
              id: 'corpus1',
              types: {
                type1: {
                  id: 'typeid',
                  slug: 'type1',
                  display_name: 'Type 1'
                }
              }
            }
          }
        }

        mutations.updateType(state, {
          corpus: 'corpus1',
          id: 'newtypeid',
          slug: 'type2',
          display_name: 'Type 2'
        })

        assert.deepStrictEqual(state, {
          corpora: {
            corpus1: {
              id: 'corpus1',
              types: {
                type1: {
                  id: 'typeid',
                  slug: 'type1',
                  display_name: 'Type 1'
                },
                type2: {
                  id: 'newtypeid',
                  slug: 'type2',
                  display_name: 'Type 2'
                }
              }
            }
          }
        })
      })

      it('causes an Error when a corpus does not exist', () => {
        const state = {
          corpora: {}
        }

        assert.throws(
          () => mutations.updateType(state, {
            corpus: 'corpus2',
            slug: 'type1',
            id: 'typeid',
            display_name: 'something'
          }),
          'Corpus corpus2 does not exist'
        )
      })
    })

    describe('addDefaultCorpus', () => {
      it('create a default corpus in the store', () => {
        const state = {
          corpora: {
            corpus1: { id: 'corpus1' }
          }
        }
        mutations.addDefaultCorpus(state, { id: 'corpus2', name: 'Second corpus' })
        assert.deepStrictEqual(state, {
          corpora: {
            corpus1: { id: 'corpus1' },
            corpus2: {
              id: 'corpus2',
              name: 'Second corpus',
              authorized_users: 1,
              created: '2000-01-01T01:00:00.000000Z',
              description: '',
              public: false,
              rights: ['read', 'write', 'admin'],
              types: [{
                display_name: 'Volume',
                folder: true,
                slug: 'volume'
              }, {
                display_name: 'Page',
                folder: false,
                slug: 'page'
              }, {
                display_name: 'Text line',
                folder: false,
                slug: 'text_line'
              }, {
                display_name: 'Paragraph',
                folder: false,
                slug: 'paragraph'
              }]
            }
          }
        })
      })
    })

    describe('remove', () => {
      it('removes an existing corpus', () => {
        const state = {
          corpora: {
            corpus1: {
              id: 'corpus1',
              name: 'Corpus 1'
            }
          }
        }

        mutations.remove(state, 'corpus1')

        assert.deepStrictEqual(state, { corpora: {} })
      })

      it('causes an Error when a corpus does not exist', () => {
        const state = {
          corpora: {
            corpus1: {
              id: 'corpus1',
              name: 'Corpus 1'
            }
          }
        }

        assert.throws(
          () => mutations.remove(state, 'corpus2'),
          'Corpus corpus2 does not exist'
        )
      })
    })

    describe('setCorpusAllowedMetadata', () => {
      it('updates allowed metadata on a corpus', () => {
        const state = {
          corpusAllowedMetadata: {
            corpus1: [
              { type: 'A', name: 'A' },
              { type: 'B', name: 'B' }
            ]
          }
        }
        const addedAllowedMeta = [
          { type: 'A', name: 'A' },
          { type: 'A', name: 'B' },
          { type: 'B', name: 'A' }
        ]

        mutations.setCorpusAllowedMetadata(state, { corpusId: 'corpus1', results: addedAllowedMeta })

        assert.deepStrictEqual(state, {
          corpusAllowedMetadata: {
            corpus1: [
              { type: 'A', name: 'A' },
              { type: 'B', name: 'B' },
              { type: 'A', name: 'B' },
              { type: 'B', name: 'A' }
            ]
          }
        })
      })
    })

    describe('updateCorpusAllowedMetadata', () => {
      it('updates an existing allowed metadata (name, type) couple in state.store.corpusAllowedMetadata[corpusId]', async () => {
        const state = {
          corpora: {},
          corporaLoaded: false,
          corpusAllowedMetadata: {
            someCorpus: [
              { id: 'oneId', name: 'oneName', type: 'oneType' },
              { id: 'twoId', name: 'twoName', type: 'twoType' },
              { id: 'threeId', name: 'threeName', type: 'threeType' }
            ]
          }
        }
        const expected = {
          corpora: {},
          corporaLoaded: false,
          corpusAllowedMetadata: {
            someCorpus: [
              { id: 'oneId', name: 'oneName', type: 'oneType' },
              { id: 'twoId', name: 'twoName', type: 'twoType' },
              { id: 'threeId', name: 'fourName', type: 'threeType' }
            ]
          }
        }
        mutations.updateCorpusAllowedMetadata(state, { corpusId: 'someCorpus', data: { id: 'threeId', type: 'threeType', name: 'fourName' } })
        assert.deepStrictEqual(state, expected)
      })
      it('throws an error if the metadata id is not found in the corpus allowed metadata', async () => {
        const state = {
          corpora: {},
          corporaLoaded: false,
          corpusAllowedMetadata: {
            someCorpus: [
              { id: 'oneId', name: 'oneName', type: 'oneType' },
              { id: 'twoId', name: 'twoName', type: 'twoType' },
              { id: 'threeId', name: 'threeName', type: 'threeType' }
            ]
          }
        }
        assert.throws(
          () => mutations.updateCorpusAllowedMetadata(state, { corpusId: 'someCorpus', data: { id: 'missingId', name: 'someName', type: 'someType' } }),
          'Allowed metadata missingId not found in corpus someCorpus'
        )
      })
      it('throws an error if there is no state.store.corpusAllowedMetadata[corpusId]', async () => {
        const state = {
          corpora: {},
          corporaLoaded: false,
          corpusAllowedMetadata: {}
        }
        assert.throws(
          () => mutations.updateCorpusAllowedMetadata(state, { corpusId: 'someCorpus', mdId: 'missingId' }),
          'Allowed metadata for corpus someCorpus not found'
        )
      })
    })

    describe('removeCorpusAllowedMetadata', () => {
      it('removes an allowed metadata from state.store.corpusAllowedMetadata[corpusId]', async () => {
        const state = {
          corpora: {},
          corporaLoaded: false,
          corpusAllowedMetadata: {
            someCorpus: [
              { id: 'oneId', name: 'oneName', type: 'oneType' },
              { id: 'twoId', name: 'twoName', type: 'twoType' },
              { id: 'threeId', name: 'threeName', type: 'threeType' }
            ]
          }
        }
        const expected = {
          corpora: {},
          corporaLoaded: false,
          corpusAllowedMetadata: {
            someCorpus: [
              { id: 'oneId', name: 'oneName', type: 'oneType' },
              { id: 'threeId', name: 'threeName', type: 'threeType' }
            ]
          }
        }
        mutations.removeCorpusAllowedMetadata(state, { corpusId: 'someCorpus', mdId: 'twoId' })
        assert.deepStrictEqual(state, expected)
      })
      it('throws an error if the metadata id is not found in the corpus allowed metadata', async () => {
        const state = {
          corpora: {},
          corporaLoaded: false,
          corpusAllowedMetadata: {
            someCorpus: [
              { id: 'oneId', name: 'oneName', type: 'oneType' },
              { id: 'twoId', name: 'twoName', type: 'twoType' },
              { id: 'threeId', name: 'threeName', type: 'threeType' }
            ]
          }
        }
        const expected = {
          corpora: {},
          corporaLoaded: false,
          corpusAllowedMetadata: {
            someCorpus: [
              { id: 'oneId', name: 'oneName', type: 'oneType' },
              { id: 'twoId', name: 'twoName', type: 'twoType' },
              { id: 'threeId', name: 'threeName', type: 'threeType' }
            ]
          }
        }
        assert.throws(
          () => mutations.removeCorpusAllowedMetadata(state, { corpusId: 'someCorpus', mdId: 'missingId' }),
          'Allowed metadata missingId not found in corpus someCorpus'
        )
        assert.deepStrictEqual(state, expected)
      })
      it('handles errors', async () => {
        const state = {
          corpora: {},
          corporaLoaded: false,
          corpusAllowedMetadata: {}
        }
        assert.throws(
          () => mutations.removeCorpusAllowedMetadata(state, { corpusId: 'someCorpus', mdId: 'oneId' }),
          'Allowed metadata for corpus someCorpus not found'
        )
      })
    })

    describe('setCorpusEntityTypes', () => {
      it('updates entity types on a corpus', () => {
        const state = {
          corpusEntityTypes: {
            corpus1: [
              { id: 'id1', name: 'type1', color: 'ff0000' },
              { id: 'id2', name: 'second type', color: '170dd9' }
            ]
          }
        }
        const newEntityTypes = [
          { id: 'id1', name: 'type1', color: 'ff0000' },
          { id: 'id3', name: 'type number three', color: '180cf7' },
          { id: 'id4', name: '4th type', color: 'f70cbd' }
        ]

        mutations.setCorpusEntityTypes(state, { corpusId: 'corpus1', results: newEntityTypes })

        assert.deepStrictEqual(state, {
          corpusEntityTypes: {
            corpus1: [
              { id: 'id1', name: 'type1', color: 'ff0000' },
              { id: 'id2', name: 'second type', color: '170dd9' },
              { id: 'id3', name: 'type number three', color: '180cf7' },
              { id: 'id4', name: '4th type', color: 'f70cbd' }
            ]
          }
        })
      })
    })

    describe('updateCorpusEntityType', () => {
      it('updates an Entity Type in state.store.corpusEntityTypes[corpusId]', async () => {
        const state = {
          corpora: {},
          corporaLoaded: false,
          corpusEntityTypes: {
            corpus1: [
              { id: 'id1', name: 'type1', color: 'ff0000' },
              { id: 'id2', name: 'second type', color: '170dd9' }
            ]
          }
        }
        const expected = {
          corpora: {},
          corporaLoaded: false,
          corpusEntityTypes: {
            corpus1: [
              { id: 'id1', name: 'first type', color: 'ff0000' },
              { id: 'id2', name: 'second type', color: '170dd9' }
            ]
          }
        }
        mutations.updateCorpusEntityType(state, { corpusId: 'corpus1', data: { id: 'id1', name: 'first type', color: 'ff0000' } })
        assert.deepStrictEqual(state, expected)
      })
      it('throws an error if the type id is not found in the corpus Entity Types', async () => {
        const state = {
          corpora: {},
          corporaLoaded: false,
          corpusEntityTypes: {
            corpus1: [
              { id: 'id1', name: 'type1', color: 'ff0000' },
              { id: 'id2', name: 'second type', color: '170dd9' }
            ]
          }
        }
        assert.throws(
          () => mutations.updateCorpusEntityType(state, { corpusId: 'corpus1', data: { id: 'id3', name: 'a name', color: 'ffffff' } }),
          'Entity Type id3 not found in corpus corpus1'
        )
      })
      it('throws an error if there is no state.store.corpusEntityTypes[corpusId]', async () => {
        const state = {
          corpora: {},
          corporaLoaded: false,
          corpusEntityTypes: {}
        }
        assert.throws(
          () => mutations.updateCorpusEntityType(state, { corpusId: 'corpus1', data: { id: 'id1', name: 'first type', color: 'ff0000' } }),
          'Entity Types for corpus corpus1 not found'
        )
      })
    })

    describe('removeCorpusEntityType', () => {
      it('removes an Entity Type from state.store.corpusEntityTypes[corpusId]', async () => {
        const state = {
          corpora: {},
          corporaLoaded: false,
          corpusEntityTypes: {
            corpus1: [
              { id: 'id1', name: 'type1', color: 'ff0000' },
              { id: 'id2', name: 'second type', color: '170dd9' },
              { id: 'id3', name: 'type number three', color: '180cf7' },
              { id: 'id4', name: '4th type', color: 'f70cbd' }
            ]
          }
        }
        const expected = {
          corpora: {},
          corporaLoaded: false,
          corpusEntityTypes: {
            corpus1: [
              { id: 'id1', name: 'type1', color: 'ff0000' },
              { id: 'id2', name: 'second type', color: '170dd9' },
              { id: 'id4', name: '4th type', color: 'f70cbd' }
            ]
          }
        }
        mutations.removeCorpusEntityType(state, { corpusId: 'corpus1', typeId: 'id3' })
        assert.deepStrictEqual(state, expected)
      })
      it('throws an error if the type id is not found in the corpus Entity Types', async () => {
        const state = {
          corpora: {},
          corporaLoaded: false,
          corpusEntityTypes: {
            corpus1: [
              { id: 'id1', name: 'type1', color: 'ff0000' },
              { id: 'id2', name: 'second type', color: '170dd9' },
              { id: 'id3', name: 'type number three', color: '180cf7' },
              { id: 'id4', name: '4th type', color: 'f70cbd' }
            ]
          }
        }
        const expected = {
          corpora: {},
          corporaLoaded: false,
          corpusEntityTypes: {
            corpus1: [
              { id: 'id1', name: 'type1', color: 'ff0000' },
              { id: 'id2', name: 'second type', color: '170dd9' },
              { id: 'id3', name: 'type number three', color: '180cf7' },
              { id: 'id4', name: '4th type', color: 'f70cbd' }
            ]
          }
        }
        assert.throws(
          () => mutations.removeCorpusEntityType(state, { corpusId: 'corpus1', typeId: 'id5' }),
          'Entity Type id5 not found in corpus corpus1'
        )
        assert.deepStrictEqual(state, expected)
      })
      it('handles errors', async () => {
        const state = {
          corpora: {},
          corporaLoaded: false,
          corpusEntityTypes: {}
        }
        assert.throws(
          () => mutations.removeCorpusEntityType(state, { corpusId: 'corpus1', mdId: 'oneId' }),
          'Entity Types for corpus corpus1 not found'
        )
      })
    })

    describe('setCorpusDatasets', () => {
      it('sets datasets on a corpus', () => {
        const state = {
          corpusDatasets: {
            corpus1: [
              {
                id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
                state: 'open',
                name: 'Shin Seiki Evangelion',
                description: 'Become legend, young boy!',
                sets: ['unit-00', 'unit-01', 'unit-02'],
                corpus_id: 'corpus1'
              }
            ]
          }
        }
        const newDataset = [
          {
            id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
            state: 'open',
            name: 'Mobile Suit Gundam Wing',
            description: 'another dataset',
            sets: ['training', 'test', 'validation']
          }
        ]

        mutations.setCorpusDatasets(state, { corpusId: 'corpus1', results: newDataset })

        assert.deepStrictEqual(state, {
          corpusDatasets: {
            corpus1: [
              {
                id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
                state: 'open',
                name: 'Shin Seiki Evangelion',
                description: 'Become legend, young boy!',
                sets: ['unit-00', 'unit-01', 'unit-02'],
                corpus_id: 'corpus1'
              },
              {
                id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
                state: 'open',
                name: 'Mobile Suit Gundam Wing',
                description: 'another dataset',
                sets: ['training', 'test', 'validation'],
                corpus_id: 'corpus1'
              }
            ]
          }
        })
      })
    })

    describe('updateCorpusDatasets', () => {
      it('updates datasets on a corpus', () => {
        const state = {
          corpusDatasets: {
            corpus1: [
              {
                id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
                state: 'open',
                name: 'Shin Seiki Evangelion',
                description: 'Become legend, young boy!',
                sets: ['unit-00', 'unit-01', 'unit-02'],
                corpus_id: 'corpus1'
              }
            ]
          }
        }
        const updatedDataset = {
          id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
          state: 'open',
          name: 'Human Instrumentality Project',
          description: 'Become legend, young boy!',
          sets: ['unit-00', 'unit-01', 'unit-02', 'unit-03']
        }

        mutations.updateCorpusDatasets(state, { corpusId: 'corpus1', data: updatedDataset })

        assert.deepStrictEqual(state, {
          corpusDatasets: {
            corpus1: [
              {
                id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
                state: 'open',
                name: 'Human Instrumentality Project',
                description: 'Become legend, young boy!',
                sets: ['unit-00', 'unit-01', 'unit-02', 'unit-03'],
                corpus_id: 'corpus1'
              }
            ]
          }
        })
      })

      it('throws an error if the dataset is not found in corpusDatasets[corpusId]', () => {
        const state = {
          corpusDatasets: {
            corpus1: [
              {
                id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
                state: 'open',
                name: 'Shin Seiki Evangelion',
                description: 'Become legend, young boy!',
                sets: ['unit-00', 'unit-01', 'unit-02'],
                corpus_id: 'corpus1'
              }
            ]
          }
        }
        const originalState = { ...state }
        const updatedDataset = {
          id: 'aaaabaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
          state: 'open',
          name: 'Human Instrumentality Project',
          description: 'Become legend, young boy!',
          sets: ['unit-00', 'unit-01', 'unit-02', 'unit-03']
        }

        assert.throws(
          () => mutations.updateCorpusDatasets(state, { corpusId: 'corpus1', data: updatedDataset }),
          'Dataset aaaabaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa not found in corpus corpus1'
        )
        assert.deepStrictEqual(state, originalState)
      })

      it('throws an error if the corpusId is not found in corpusDatasets', () => {
        const state = {
          corpusDatasets: {
            corpus1: [
              {
                id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
                state: 'open',
                name: 'Shin Seiki Evangelion',
                description: 'Become legend, young boy!',
                sets: ['unit-00', 'unit-01', 'unit-02']
              }
            ]
          }
        }
        const originalState = { ...state }
        const updatedDataset = {
          id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
          state: 'open',
          name: 'Human Instrumentality Project',
          description: 'Become legend, young boy!',
          sets: ['unit-00', 'unit-01', 'unit-02', 'unit-03']
        }

        assert.throws(
          () => mutations.updateCorpusDatasets(state, { corpusId: 'corpus2', data: updatedDataset }),
          'Datasets for corpus corpus2 not found'
        )
        assert.deepStrictEqual(state, originalState)
      })
    })

    describe('removeCorpusDataset', () => {
      it('removes a dataset from a corpus', () => {
        const state = {
          corpusDatasets: {
            corpus1: [
              {
                id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
                state: 'open',
                name: 'Shin Seiki Evangelion',
                description: 'Become legend, young boy!',
                sets: ['unit-00', 'unit-01', 'unit-02'],
                corpus_id: 'corpus1'
              },
              {
                id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
                state: 'open',
                name: 'Mobile Suit Gundam Wing',
                description: 'another dataset',
                sets: ['training', 'test', 'validation'],
                corpus_id: 'corpus1'
              }
            ]
          }
        }

        mutations.removeCorpusDataset(state, { corpusId: 'corpus1', datasetId: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa' })

        assert.deepStrictEqual(state, {
          corpusDatasets: {
            corpus1: [
              {
                id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
                state: 'open',
                name: 'Mobile Suit Gundam Wing',
                description: 'another dataset',
                sets: ['training', 'test', 'validation'],
                corpus_id: 'corpus1'
              }
            ]
          }
        })
      })

      it('throws an error if the dataset is not found in corpusDatasets[corpusId]', () => {
        const state = {
          corpusDatasets: {
            corpus1: [
              {
                id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
                state: 'open',
                name: 'Shin Seiki Evangelion',
                description: 'Become legend, young boy!',
                sets: ['unit-00', 'unit-01', 'unit-02']
              }
            ]
          }
        }
        const originalState = { ...state }

        assert.throws(
          () => mutations.removeCorpusDataset(state, { corpusId: 'corpus1', datasetId: 'aaaaabaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa' }),
          'Dataset aaaaabaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa not found in corpus corpus1'
        )
        assert.deepStrictEqual(state, originalState)
      })

      it('throws an error if the corpusId is not found in corpusDatasets', () => {
        const state = {
          corpusDatasets: {
            corpus1: [
              {
                id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
                state: 'open',
                name: 'Shin Seiki Evangelion',
                description: 'Become legend, young boy!',
                sets: ['unit-00', 'unit-01', 'unit-02']
              }
            ]
          }
        }
        const originalState = { ...state }

        assert.throws(
          () => mutations.removeCorpusDataset(state, { corpusId: 'corpus2', data: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa' }),
          'Datasets for corpus corpus2 not found'
        )
        assert.deepStrictEqual(state, originalState)
      })
    })

    it('setExports', () => {
      const state = { exports: null }
      mutations.setExports(state, { count: 1, results: [exportSample] })
      assert.deepStrictEqual(state, {
        exports: {
          count: 1,
          results: [exportSample]
        }
      })
    })

    it('reset', () => {
      const state = {
        corpora: {
          corpusid: { id: 'corpusid' }
        },
        corporaLoaded: true,
        exports: {
          count: 4
        }
      }
      mutations.reset(state)
      assert.deepStrictEqual(state, initialState())
    })
  })

  describe('actions', () => {
    let mock, jobsStore

    before('Setting up mocks', () => {
      mock = new FakeAxios(axios)
      setActivePinia(createPinia())
      jobsStore = useJobsStore()
    })

    afterEach(() => {
      // Remove any handlers, but leave mocking in place
      mock.reset()
      store.reset()
      jobsStore.$reset()
    })

    after('Removing Axios mock', () => {
      // Remove mocking entirely
      mock.restore()
    })

    describe('list', () => {
      it('loads corpora and returns them', async () => {
        const reply = [
          {
            id: 'corpus1',
            name: 'Corpus 1',
            rights: ['read'],
            types: []
          }
        ]
        mock.onGet('/corpus/').reply(200, reply)

        const corpora = await store.dispatch('corpora/list')

        assert.deepStrictEqual(store.history, [
          { action: 'corpora/list' },
          { action: 'corpora/$_fetch' },
          { mutation: 'corpora/set', payload: reply }
        ])

        assert.deepStrictEqual(store.state.corpora.corpora, corpora)
        assert.strictEqual(store.state.corpora.corporaLoaded, true)
        assert.deepStrictEqual(corpora, {
          corpus1: {
            id: 'corpus1',
            name: 'Corpus 1',
            rights: ['read'],
            types: {}
          }
        })
      })

      it('loads corpora only when they are not loaded', async () => {
        store.setState('corpora.corpora', {
          corpus1: {
            id: 'corpus1',
            name: 'Corpus 1',
            rights: ['read'],
            types: {}
          }
        })

        const corpora = await store.dispatch('corpora/list')

        assert.deepStrictEqual(store.state.corpora.corpora, corpora)
        assert.deepStrictEqual(store.history, [
          { action: 'corpora/list' }
        ])
      })

      it('loads corpora only when they are not loaded, even if there are none', async () => {
        store.state.corpora.corpora = {}
        store.state.corpora.corporaLoaded = true

        const corpora = await store.dispatch('corpora/list')

        assert.deepStrictEqual(store.state.corpora.corpora, corpora)
        assert.strictEqual(store.state.corpora.corporaLoaded, true)
        assert.deepStrictEqual(store.history, [
          { action: 'corpora/list' }
        ])
      })
    })

    describe('get', () => {
      it('loads corpora and returns one', async () => {
        const reply = [
          {
            id: 'corpus1',
            name: 'Corpus 1',
            rights: ['read'],
            types: []
          },
          {
            id: 'corpus2',
            name: 'Corpus 2',
            rights: ['read'],
            types: []
          }
        ]
        mock.onGet('/corpus/').reply(200, reply)

        const corpus = await store.dispatch('corpora/get', { id: 'corpus1' })

        assert.deepStrictEqual(store.history, [
          { action: 'corpora/get', payload: { id: 'corpus1' } },
          { action: 'corpora/$_fetch' },
          { mutation: 'corpora/set', payload: reply }
        ])

        assert.deepStrictEqual(corpus, {
          id: 'corpus1',
          name: 'Corpus 1',
          rights: ['read'],
          types: {}
        })
      })

      it('loads corpora only when they are not loaded', async () => {
        store.setState('corpora.corpora', {
          corpus1: {
            id: 'corpus1',
            name: 'Corpus 1',
            rights: ['read'],
            types: {}
          },
          corpus2: {
            id: 'corpus2',
            name: 'Corpus 2',
            rights: ['read'],
            types: {}
          }
        })

        const corpus = await store.dispatch('corpora/get', { id: 'corpus1' })

        assert.deepStrictEqual(store.history, [
          { action: 'corpora/get', payload: { id: 'corpus1' } }
        ])

        assert.deepStrictEqual(corpus, {
          id: 'corpus1',
          name: 'Corpus 1',
          rights: ['read'],
          types: {}
        })
      })

      it('causes an Error when a corpus does not exist', async () => {
        store.setState('corpora.corpora', { corpus1: { id: 'corpus1' } })
        const error = await assertRejects(async () => store.dispatch('corpora/get', { id: 'corpus42' }))
        assert.strictEqual(error.toString(), 'Error: Corpus with ID corpus42 not found')
      })
    })

    describe('create', () => {
      it('creates a new corpus', async () => {
        const reply = {
          id: 'corpus42',
          name: 'New corpus',
          types: []
        }
        mock.onPost('/corpus/').reply(201, reply)

        store.setState('corpora.corpora', {
          corpus1: { id: 'corpus1' }
        })

        await store.dispatch('corpora/create', { name: 'New corpus' })

        assert.deepStrictEqual(store.history, [
          { action: 'corpora/create', payload: { name: 'New corpus' } },
          { mutation: 'corpora/update', payload: reply }
        ])

        assert.deepStrictEqual(store.state.corpora.corpora, {
          corpus1: { id: 'corpus1' },
          corpus42: { id: 'corpus42', name: 'New corpus', types: {} }
        })
      })

      it('loads corpora before creating', async () => {
        const listReply = [
          {
            id: 'corpus1',
            name: 'Corpus 1',
            types: []
          }
        ]
        const createReply = {
          id: 'corpus42',
          name: 'New corpus',
          types: []
        }
        mock.onGet('/corpus/').reply(200, listReply)
        mock.onPost('/corpus/').reply(201, createReply)

        await store.dispatch('corpora/create', { name: 'New corpus' })

        assert.deepStrictEqual(store.history, [
          { action: 'corpora/create', payload: { name: 'New corpus' } },
          { action: 'corpora/$_fetch' },
          { mutation: 'corpora/set', payload: listReply },
          { mutation: 'corpora/update', payload: createReply }
        ])

        assert.deepStrictEqual(store.state.corpora.corpora, {
          corpus1: { id: 'corpus1', name: 'Corpus 1', types: {} },
          corpus42: { id: 'corpus42', name: 'New corpus', types: {} }
        })
      })

      it('throws errors', async () => {
        mock.onPost('/corpus/').reply(403, { detail: ['You must be logged in.'] })
        store.setState('corpora.corpora', { corpus1: { id: 'corpus1' } })
        const error = await assertRejects(async () => store.dispatch('corpora/create', { name: 'New corpus' }))
        assert.deepStrictEqual(error.response.data, { detail: ['You must be logged in.'] })
      })
    })

    describe('update', () => {
      it('updates an existing corpus', async () => {
        const reply = {
          id: 'corpus1',
          name: 'New name',
          types: []
        }
        mock.onPatch('/corpus/corpus1/').reply(201, reply)

        store.setState('corpora.corpora', {
          corpus1: { id: 'corpus1', name: 'Old name' }
        })

        await store.dispatch('corpora/update', { id: 'corpus1', name: 'New name' })

        assert.deepStrictEqual(store.history, [
          { action: 'corpora/update', payload: { id: 'corpus1', name: 'New name' } },
          { mutation: 'corpora/update', payload: reply }
        ])

        assert.deepStrictEqual(store.state.corpora.corpora, {
          corpus1: { id: 'corpus1', name: 'New name', types: {} }
        })
      })

      it('loads corpora before updating', async () => {
        const listReply = [
          {
            id: 'corpus1',
            name: 'Old name',
            types: []
          }
        ]
        const updateReply = {
          id: 'corpus1',
          name: 'New name',
          types: []
        }
        mock.onGet('/corpus/').reply(200, listReply)
        mock.onPatch('/corpus/corpus1/').reply(201, updateReply)

        await store.dispatch('corpora/update', { id: 'corpus1', name: 'New name' })

        assert.deepStrictEqual(store.history, [
          { action: 'corpora/update', payload: { id: 'corpus1', name: 'New name' } },
          { action: 'corpora/$_fetch' },
          { mutation: 'corpora/set', payload: listReply },
          { mutation: 'corpora/update', payload: updateReply }
        ])

        assert.deepStrictEqual(store.state.corpora.corpora, {
          corpus1: { id: 'corpus1', name: 'New name', types: {} }
        })
      })

      it('throws errors', async () => {
        mock.onPatch('/corpus/corpus1/').reply(403, { detail: ['You must be logged in.'] })
        store.setState('corpora.corpora', { corpus1: { id: 'corpus1' } })
        const error = await assertRejects(async () => store.dispatch('corpora/update', { id: 'corpus1', name: 'New name' }))
        assert.deepStrictEqual(error.response.data, { detail: ['You must be logged in.'] })
      })
    })

    describe('delete', () => {
      it('deletes an existing corpus', async () => {
        mock.onDelete('/corpus/corpus1/').reply(204)
        mock.onGet('/jobs/').reply(200, jobsSample)

        store.setState('corpora.corpora', {
          corpus1: { id: 'corpus1', name: 'Corpus 1' }
        })

        await store.dispatch('corpora/delete', { id: 'corpus1' })
        await store.actionsCompleted()

        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/delete',
            payload: { id: 'corpus1' }
          },
          {
            mutation: 'corpora/remove',
            payload: 'corpus1'
          }
        ])

        assert.deepStrictEqual(store.state.corpora.corpora, {})
        assert.deepStrictEqual(jobsStore.jobs, {
          jobid: jobsSample[0]
        })
      })

      it('loads corpora before deleting', async () => {
        const reply = [
          {
            id: 'corpus1',
            name: 'Corpus 1',
            types: []
          }
        ]
        mock.onGet('/corpus/').reply(200, reply)
        mock.onDelete('/corpus/corpus1/').reply(204)
        mock.onGet('/jobs/').reply(200, jobsSample)

        await store.dispatch('corpora/delete', { id: 'corpus1' })
        await store.actionsCompleted()

        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/delete',
            payload: { id: 'corpus1' }
          },
          {
            action: 'corpora/$_fetch'
          },
          {
            mutation: 'corpora/set',
            payload: reply
          },
          {
            mutation: 'corpora/remove',
            payload: 'corpus1'
          }
        ])

        assert.deepStrictEqual(store.state.corpora.corpora, {})
        assert.deepStrictEqual(jobsStore.jobs, {
          jobid: jobsSample[0]
        })
      })

      it('throws parsed errors', async () => {
        mock.onDelete('/corpus/corpus1/').reply(403, { detail: ['You must be logged in.'] })
        mock.onGet('/jobs/').reply(200, jobsSample)
        store.state.corpora.corpora = { corpus1: { id: 'corpus1' } }

        const error = await assertRejects(store.dispatch, 'corpora/delete', { id: 'corpus1' })
        await store.actionsCompleted()

        assert.strictEqual(error.toString(), 'Error: You must be logged in.')

        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/delete',
            payload: { id: 'corpus1' }
          }
        ])
        assert.deepStrictEqual(store.state.corpora.corpora, { corpus1: { id: 'corpus1' } })
        assert.deepStrictEqual(jobsStore.jobs, {
          jobid: jobsSample[0]
        })
      })
    })

    describe('getType', () => {
      it('loads corpora and returns a type', async () => {
        const reply = [
          {
            id: 'corpus1',
            name: 'Corpus 1',
            rights: ['read', 'write'],
            types: [
              {
                slug: 'type1',
                display_name: 'Type 1'
              },
              {
                slug: 'type2',
                display_name: 'Type 2'
              }
            ]
          }
        ]

        mock.onGet('/corpus/').reply(200, reply)

        const type = await store.dispatch('corpora/getType', { id: 'corpus1', slug: 'type2' })

        assert.deepStrictEqual(store.history, [
          { action: 'corpora/getType', payload: { id: 'corpus1', slug: 'type2' } },
          { action: 'corpora/get', payload: { id: 'corpus1' } },
          { action: 'corpora/$_fetch' },
          { mutation: 'corpora/set', payload: reply }
        ])

        assert.deepStrictEqual(type, {
          slug: 'type2',
          display_name: 'Type 2'
        })
      })

      it('loads corpora only when they are not loaded', async () => {
        store.setState('corpora.corpora', {
          corpus1: {
            id: 'corpus1',
            name: 'Corpus 1',
            types: {
              type1: {
                slug: 'type1',
                display_name: 'Type 1'
              },
              type2: {
                slug: 'type2',
                display_name: 'Type 2'
              }
            }
          }
        })

        const type = await store.dispatch('corpora/getType', { id: 'corpus1', slug: 'type2' })

        assert.deepStrictEqual(store.history, [
          { action: 'corpora/getType', payload: { id: 'corpus1', slug: 'type2' } },
          { action: 'corpora/get', payload: { id: 'corpus1' } }
        ])

        assert.deepStrictEqual(type, {
          slug: 'type2',
          display_name: 'Type 2'
        })
      })

      it('causes an Error when a corpus does not exist', async () => {
        store.setState('corpora.corpora', {
          corpus1: {
            id: 'corpus1',
            name: 'Corpus 1',
            types: {
              type1: {
                slug: 'type1',
                display_name: 'Type 1'
              },
              type2: {
                slug: 'type2',
                display_name: 'Type 2'
              }
            }
          }
        })

        const error = await assertRejects(async () => store.dispatch('corpora/getType', { id: 'corpus42', slug: 'type3' }))

        assert.strictEqual(error.toString(), 'Error: Corpus with ID corpus42 not found')
      })

      it('causes an Error when a type does not exist', async () => {
        store.setState('corpora.corpora', {
          corpus1: {
            id: 'corpus1',
            name: 'Corpus 1',
            types: {
              type1: {
                slug: 'type1',
                display_name: 'Type 1'
              },
              type2: {
                slug: 'type2',
                display_name: 'Type 2'
              }
            }
          }
        })

        const error = await assertRejects(async () => store.dispatch('corpora/getType', { id: 'corpus1', slug: 'type3' }))

        assert.strictEqual(error.toString(), 'Error: Type type3 not found on corpus corpus1')
      })
    })

    describe('updateType', () => {
      const payload = {
        id: 'typeid',
        display_name: 'newName',
        slug: 'newSlug',
        folder: false,
        corpus: 'corpusid'
      }

      it('updates an element type', async () => {
        const response = { ...payload }
        delete response.corpus
        mock.onPatch('/elements/type/typeid/').reply(200, response)
        store.setState('corpora.corpora.corpusid', { id: 'corpusid', types: {} })

        await store.dispatch('corpora/updateType', payload)

        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/updateType',
            payload
          },
          {
            mutation: 'corpora/updateType',
            payload
          }
        ])

        assert.deepStrictEqual(store.state.corpora.corpora.corpusid.types, {
          newSlug: response
        })
      })

      it('handles errors', async () => {
        mock.onPatch('/elements/type/typeid/').reply(400)

        await store.dispatch('corpora/updateType', payload)

        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/updateType',
            payload
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 400'
            }
          }
        ])
      })
    })

    describe('createType', () => {
      const payload = {
        display_name: 'newName',
        slug: 'newSlug',
        folder: false,
        corpus: 'corpusid'
      }

      it('creates an element type', async () => {
        const response = {
          id: 'typeid',
          display_name: 'newName',
          slug: 'newSlug',
          folder: false
        }
        mock.onPost('/elements/type/').reply(201, response)
        store.setState('corpora.corpora.corpusid', { id: 'corpusid', types: {} })

        await store.dispatch('corpora/createType', payload)

        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/createType',
            payload
          },
          {
            mutation: 'corpora/updateType',
            payload: {
              corpus: 'corpusid',
              ...response
            }
          }
        ])

        assert.deepStrictEqual(store.state.corpora.corpora.corpusid.types, {
          newSlug: response
        })
      })

      it('handles errors', async () => {
        mock.onPost('/elements/type/').reply(400)

        await store.dispatch('corpora/createType', payload)

        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/createType',
            payload
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 400'
            }
          }
        ])
      })
    })

    describe('listCorpusEntityTypes', () => {
      it('lists entity types in a corpus', async () => {
        const pages = [{
          count: 4,
          number: 1,
          next: 'nextpage',
          results: [
            { id: 'id1', name: 'type1', color: 'ff0000' },
            { id: 'id2', name: 'second type', color: '170dd9' },
            { id: 'id3', name: 'type number three', color: '180cf7' }
          ]
        }, {
          count: 4,
          number: 2,
          next: null,
          results: [{ id: 'id4', name: '4th type', color: 'f70cbd' }]
        }]
        mock.onGet('/corpus/corpusid/entity-types/', { params: { page: 1 } }).reply(200, pages[0])
        mock.onGet('/corpus/corpusid/entity-types/', { params: { page: 2 } }).reply(200, pages[1])

        await store.dispatch('corpora/listCorpusEntityTypes', { corpusId: 'corpusid' })

        assert.deepStrictEqual(store.history, [
          { action: 'corpora/listCorpusEntityTypes', payload: { corpusId: 'corpusid' } },
          { mutation: 'corpora/setCorpusEntityTypes', payload: { corpusId: 'corpusid', results: pages[0].results } },
          { action: 'corpora/listCorpusEntityTypes', payload: { corpusId: 'corpusid', page: 2 } },
          { mutation: 'corpora/setCorpusEntityTypes', payload: { corpusId: 'corpusid', results: pages[1].results } }
        ])

        assert.deepStrictEqual(store.state.corpora.corpusEntityTypes, {
          corpusid: [
            { id: 'id1', name: 'type1', color: 'ff0000' },
            { id: 'id2', name: 'second type', color: '170dd9' },
            { id: 'id3', name: 'type number three', color: '180cf7' },
            { id: 'id4', name: '4th type', color: 'f70cbd' }
          ]
        })
      })
    })

    describe('createCorpusEntityType', () => {
      it('creates a new entity type in a corpus', async () => {
        store.state.corpora.corpusEntityTypes = {
          testCorpus: [
            { id: 'id1', name: 'type1', color: 'ff0000' }
          ]
        }
        const response = {
          id: 'id2',
          name: 'other type',
          color: 'f7240c'
        }
        mock.onPost('/entity/types/').reply(201, response)

        await store.dispatch('corpora/createCorpusEntityType', { corpus: 'testCorpus', name: 'other type', color: 'f7240c' })

        assert.deepStrictEqual(store.state.corpora.corpusEntityTypes, {
          testCorpus: [
            { id: 'id1', name: 'type1', color: 'ff0000' },
            { id: 'id2', name: 'other type', color: 'f7240c' }
          ]
        })
        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/createCorpusEntityType',
            payload: { corpus: 'testCorpus', name: 'other type', color: 'f7240c' }
          },
          {
            mutation: 'corpora/setCorpusEntityTypes',
            payload: { corpusId: 'testCorpus', results: [{ id: 'id2', name: 'other type', color: 'f7240c' }] }
          }
        ])
      })

      it('handles errors', async () => {
        mock.onPost('/entity/types/').reply(400)
        await assertRejects(async () => store.dispatch('corpora/createCorpusEntityType', { corpus: 'testCorpus', name: 'type1' }))

        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/createCorpusEntityType',
            payload: { corpus: 'testCorpus', name: 'type1' }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 400'
            }
          }
        ])
      })
    })

    describe('updateCorpusEntityType', () => {
      it('updates an existing Entity Type in a corpus', async () => {
        store.state.corpora.corpusEntityTypes = {
          corpus1: [
            { id: 'id1', name: 'type1', color: 'ff0000' },
            { id: 'id2', name: 'second type', color: '170dd9' },
            { id: 'id3', name: 'type number three', color: '180cf7' }
          ]
        }
        const response = { id: 'id2', name: 'another type', color: '170dd9' }
        mock.onPatch('/entity/types/id2/').reply(200, response)
        await store.dispatch('corpora/updateCorpusEntityType', { id: 'id2', name: 'another type', color: '170dd', corpusId: 'corpus1' })
        assert.deepStrictEqual(store.state.corpora.corpusEntityTypes, {
          corpus1: [
            { id: 'id1', name: 'type1', color: 'ff0000' },
            { id: 'id2', name: 'another type', color: '170dd9' },
            { id: 'id3', name: 'type number three', color: '180cf7' }
          ]
        })
        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/updateCorpusEntityType',
            payload: { id: 'id2', name: 'another type', color: '170dd', corpusId: 'corpus1' }
          },
          {
            mutation: 'corpora/updateCorpusEntityType',
            payload: { corpusId: 'corpus1', data: { id: 'id2', name: 'another type', color: '170dd9' } }
          }
        ])
      })
      it('handles errors', async () => {
        mock.onPatch('/entity/types/id5/').reply(400)
        await assertRejects(async () => await store.dispatch('corpora/updateCorpusEntityType', { id: 'id5', name: 'bonk', color: '170dd', corpusId: 'corpus1' }))
        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/updateCorpusEntityType',
            payload: { id: 'id5', name: 'bonk', color: '170dd', corpusId: 'corpus1' }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 400'
            }
          }
        ])
      })
    })

    describe('deleteCorpusEntityType', () => {
      it('deletes an Entity Type from a corpus', async () => {
        store.state.corpora.corpusEntityTypes = {
          corpus1: [
            { id: 'id1', name: 'type1', color: 'ff0000' },
            { id: 'id2', name: 'another type', color: '170dd9' },
            { id: 'id3', name: 'type number three', color: '180cf7' }
          ]
        }
        mock.onDelete('/entity/types/id2/').reply(204)
        await store.dispatch('corpora/deleteCorpusEntityType', { corpusId: 'corpus1', typeId: 'id2' })
        assert.deepStrictEqual(store.state.corpora.corpusEntityTypes, {
          corpus1: [
            { id: 'id1', name: 'type1', color: 'ff0000' },
            { id: 'id3', name: 'type number three', color: '180cf7' }
          ]
        })
        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/deleteCorpusEntityType',
            payload: { corpusId: 'corpus1', typeId: 'id2' }
          },
          {
            mutation: 'corpora/removeCorpusEntityType',
            payload: { corpusId: 'corpus1', typeId: 'id2' }
          }
        ])
      })
      it('handles errors', async () => {
        mock.onDelete('/entity/types/id5/').reply(400)
        await assertRejects(async () => await store.dispatch('corpora/deleteCorpusEntityType', { corpusId: 'corpus1', typeId: 'id5' }))
        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/deleteCorpusEntityType',
            payload: { corpusId: 'corpus1', typeId: 'id5' }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 400'
            }
          }
        ])
      })
    })

    describe('listAllowedMetadata', () => {
      it('recursively lists allowed metadata for a corpus', async () => {
        const pages = [{
          count: 3,
          number: 1,
          next: 'nexpage',
          results: [{ type: 'A', name: 'A' }, { type: 'A', name: 'B' }]
        }, {
          count: 3,
          number: 2,
          next: null,
          results: [{ type: 'B', name: 'B' }]
        }]
        mock.onGet('/corpus/corpusid/allowed-metadata/', { params: { page: 1 } }).reply(200, pages[0])
        mock.onGet('/corpus/corpusid/allowed-metadata/', { params: { page: 2 } }).reply(200, pages[1])

        await store.dispatch('corpora/listAllowedMetadata', { corpusId: 'corpusid' })

        assert.deepStrictEqual(store.history, [
          { action: 'corpora/listAllowedMetadata', payload: { corpusId: 'corpusid' } },
          { mutation: 'corpora/setCorpusAllowedMetadata', payload: { corpusId: 'corpusid', results: pages[0].results } },
          { action: 'corpora/listAllowedMetadata', payload: { corpusId: 'corpusid', page: 2 } },
          { mutation: 'corpora/setCorpusAllowedMetadata', payload: { corpusId: 'corpusid', results: pages[1].results } }
        ])

        assert.deepStrictEqual(store.state.corpora.corpusAllowedMetadata, {
          corpusid: [
            { name: 'A', type: 'A' },
            { name: 'B', type: 'A' },
            { name: 'B', type: 'B' }
          ]
        })
      })
    })

    describe('createAllowedMetadata', () => {
      it('creates a new allowed metadata in a corpus', async () => {
        store.state.corpora.corpusAllowedMetadata = {
          someCorpus: [
            { id: 'oneId', name: 'oneName', type: 'oneType' }
          ]
        }
        const response = {
          id: 'metaId',
          type: 'someType',
          name: 'someName'
        }
        mock.onPost('/corpus/someCorpus/allowed-metadata/').reply(201, response)

        await store.dispatch('corpora/createAllowedMetadata', { corpusId: 'someCorpus', type: 'someType', name: 'someName' })

        assert.deepStrictEqual(store.state.corpora.corpusAllowedMetadata, {
          someCorpus: [
            { id: 'oneId', name: 'oneName', type: 'oneType' },
            { id: 'metaId', name: 'someName', type: 'someType' }
          ]
        })
        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/createAllowedMetadata',
            payload: { corpusId: 'someCorpus', type: 'someType', name: 'someName' }
          },
          {
            mutation: 'corpora/setCorpusAllowedMetadata',
            payload: { corpusId: 'someCorpus', results: [{ id: 'metaId', type: 'someType', name: 'someName' }] }
          }
        ])
      })

      it('handles errors', async () => {
        mock.onPost('/corpus/someCorpus/allowed-metadata/').reply(400)
        await assertRejects(async () => store.dispatch('corpora/createAllowedMetadata', { corpusId: 'someCorpus', type: 'someType', name: 'someName' }))

        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/createAllowedMetadata',
            payload: { corpusId: 'someCorpus', type: 'someType', name: 'someName' }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 400'
            }
          }
        ])
      })
    })

    describe('deleteAllowedMetadata', () => {
      it('removes an allowed metadata (name, type) couple from a corpus', async () => {
        store.state.corpora.corpusAllowedMetadata = {
          someCorpus: [
            { id: 'oneId', name: 'oneName', type: 'oneType' },
            { id: 'twoId', name: 'twoName', type: 'twoType' },
            { id: 'threeId', name: 'threeName', type: 'threeType' }
          ]
        }
        mock.onDelete('/corpus/someCorpus/allowed-metadata/twoId/').reply(204)
        await store.dispatch('corpora/deleteAllowedMetadata', { corpusId: 'someCorpus', mdId: 'twoId' })
        assert.deepStrictEqual(store.state.corpora.corpusAllowedMetadata, {
          someCorpus: [
            { id: 'oneId', name: 'oneName', type: 'oneType' },
            { id: 'threeId', name: 'threeName', type: 'threeType' }
          ]
        })
        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/deleteAllowedMetadata',
            payload: { corpusId: 'someCorpus', mdId: 'twoId' }
          },
          {
            mutation: 'corpora/removeCorpusAllowedMetadata',
            payload: { corpusId: 'someCorpus', mdId: 'twoId' }
          }
        ])
      })
      it('handles errors', async () => {
        mock.onDelete('/corpus/someCorpus/allowed-metadata/missingId/').reply(400)
        await assertRejects(async () => await store.dispatch('corpora/deleteAllowedMetadata', { corpusId: 'someCorpus', mdId: 'missingId' }))
        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/deleteAllowedMetadata',
            payload: { corpusId: 'someCorpus', mdId: 'missingId' }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 400'
            }
          }
        ])
      })
    })

    describe('updateAllowedMetadata', () => {
      it('updates an existing allowed metadata (name, type) couple in a corpus', async () => {
        store.state.corpora.corpusAllowedMetadata = {
          someCorpus: [
            { id: 'oneId', name: 'oneName', type: 'oneType' },
            { id: 'twoId', name: 'twoName', type: 'twoType' },
            { id: 'threeId', name: 'threeName', type: 'threeType' }
          ]
        }
        const response = {
          id: 'threeId',
          type: 'fourType',
          name: 'threeName'
        }
        mock.onPatch('/corpus/someCorpus/allowed-metadata/threeId/').reply(200, response)
        await store.dispatch('corpora/updateAllowedMetadata', { corpusId: 'someCorpus', mdId: 'threeId', type: 'fourType', name: 'threeName' })
        assert.deepStrictEqual(store.state.corpora.corpusAllowedMetadata, {
          someCorpus: [
            { id: 'oneId', name: 'oneName', type: 'oneType' },
            { id: 'twoId', name: 'twoName', type: 'twoType' },
            { id: 'threeId', name: 'threeName', type: 'fourType' }
          ]
        })
        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/updateAllowedMetadata',
            payload: { corpusId: 'someCorpus', mdId: 'threeId', type: 'fourType', name: 'threeName' }
          },
          {
            mutation: 'corpora/updateCorpusAllowedMetadata',
            payload: { corpusId: 'someCorpus', data: { id: 'threeId', type: 'fourType', name: 'threeName' } }
          }
        ])
      })
      it('handles errors', async () => {
        mock.onPatch('/corpus/someCorpus/allowed-metadata/missingId/').reply(400)
        await assertRejects(async () => await store.dispatch('corpora/updateAllowedMetadata', { corpusId: 'someCorpus', mdId: 'missingId', name: 'someName', type: 'someType' }))
        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/updateAllowedMetadata',
            payload: { corpusId: 'someCorpus', mdId: 'missingId', name: 'someName', type: 'someType' }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 400'
            }
          }
        ])
      })
    })

    describe('listCorpusDatasets', () => {
      it('lists datasets in a corpus', async () => {
        const pages = [
          {
            count: 3,
            number: 1,
            next: 'nextpage',
            results: [
              {
                id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
                state: 'open',
                name: 'Shin Seiki Evangelion',
                description: 'Become legend, young boy!',
                sets: ['unit-00', 'unit-01', 'unit-02']
              },
              {
                id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
                state: 'open',
                name: 'Mobile Suit Gundam Wing',
                description: 'another dataset',
                sets: ['training', 'test', 'validation']
              }
            ]
          },
          {
            count: 3,
            number: 2,
            next: null,
            results: [
              {
                id: 'cccccccc-cccc-cccc-cccc-cccccccccccc',
                state: 'open',
                name: 'Sailor Moon',
                description: 'Moon prism power, make up!',
                sets: ['neptune', 'uranus', 'pluto', 'saturn']
              }
            ]
          }
        ]
        mock.onGet('/corpus/corpus1/datasets/', { params: { page: 1 } }).reply(200, pages[0])
        mock.onGet('/corpus/corpus1/datasets/', { params: { page: 2 } }).reply(200, pages[1])

        await store.dispatch('corpora/listCorpusDatasets', { corpusId: 'corpus1' })

        assert.deepStrictEqual(store.history, [
          { action: 'corpora/listCorpusDatasets', payload: { corpusId: 'corpus1' } },
          { mutation: 'corpora/setCorpusDatasets', payload: { corpusId: 'corpus1', results: pages[0].results } },
          { action: 'corpora/listCorpusDatasets', payload: { corpusId: 'corpus1', page: 2 } },
          { mutation: 'corpora/setCorpusDatasets', payload: { corpusId: 'corpus1', results: pages[1].results } }
        ])

        assert.deepStrictEqual(store.state.corpora.corpusDatasets, {
          corpus1: [
            {
              id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
              state: 'open',
              name: 'Shin Seiki Evangelion',
              description: 'Become legend, young boy!',
              sets: ['unit-00', 'unit-01', 'unit-02'],
              corpus_id: 'corpus1'
            },
            {
              id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
              state: 'open',
              name: 'Mobile Suit Gundam Wing',
              description: 'another dataset',
              sets: ['training', 'test', 'validation'],
              corpus_id: 'corpus1'
            },
            {
              id: 'cccccccc-cccc-cccc-cccc-cccccccccccc',
              state: 'open',
              name: 'Sailor Moon',
              description: 'Moon prism power, make up!',
              sets: ['neptune', 'uranus', 'pluto', 'saturn'],
              corpus_id: 'corpus1'
            }
          ]
        })
      })
    })

    describe('createCorpusDataset', () => {
      it('creates a new dataset in a corpus', async () => {
        store.state.corpora.corpusDatasets = {
          corpus1: [
            {
              id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
              state: 'open',
              name: 'Shin Seiki Evangelion',
              description: 'Become legend, young boy!',
              sets: ['unit-00', 'unit-01', 'unit-02'],
              corpus_id: 'corpus1'
            },
            {
              id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
              state: 'open',
              name: 'Mobile Suit Gundam Wing',
              description: 'another dataset',
              sets: ['training', 'test', 'validation'],
              corpus_id: 'corpus1'
            }
          ]
        }
        const requestData = {
          name: 'Sailor Moon',
          description: 'Moon prism power, make up!',
          sets: ['neptune', 'uranus', 'pluto', 'saturn']
        }
        const responseData = {
          ...requestData,
          id: 'cccccccc-cccc-cccc-cccc-cccccccccccc',
          state: 'open'
        }
        mock.onPost('/corpus/corpus1/datasets/', { ...requestData }).reply(201, { ...responseData })

        await store.dispatch('corpora/createCorpusDataset', { corpusId: 'corpus1', ...requestData })

        assert.deepStrictEqual(store.history, [
          { action: 'corpora/createCorpusDataset', payload: { corpusId: 'corpus1', ...requestData } },
          { mutation: 'corpora/setCorpusDatasets', payload: { corpusId: 'corpus1', results: [responseData] } }
        ])

        assert.deepStrictEqual(store.state.corpora.corpusDatasets, {
          corpus1: [
            {
              id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
              state: 'open',
              name: 'Shin Seiki Evangelion',
              description: 'Become legend, young boy!',
              sets: ['unit-00', 'unit-01', 'unit-02'],
              corpus_id: 'corpus1'
            },
            {
              id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
              state: 'open',
              name: 'Mobile Suit Gundam Wing',
              description: 'another dataset',
              sets: ['training', 'test', 'validation'],
              corpus_id: 'corpus1'
            },
            {
              id: 'cccccccc-cccc-cccc-cccc-cccccccccccc',
              state: 'open',
              name: 'Sailor Moon',
              description: 'Moon prism power, make up!',
              sets: ['neptune', 'uranus', 'pluto', 'saturn'],
              corpus_id: 'corpus1'
            }
          ]
        })
      })

      it('handles errors', async () => {
        const requestData = {
          name: 'Sailor Moon',
          description: 'Moon prism power, make up!',
          sets: ['neptune', 'uranus', 'pluto', 'saturn']
        }
        mock.onPost('/corpus/corpus1/datasets/').reply(400)
        await assertRejects(async () => store.dispatch('corpora/createCorpusDataset', { corpusId: 'corpus1', ...requestData }))

        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/createCorpusDataset', payload: { corpusId: 'corpus1', ...requestData }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 400'
            }
          }
        ])
      })
    })

    describe('updateCorpusDataset', () => {
      it('updates an existing dataset in a corpus', async () => {
        store.state.corpora.corpusDatasets = {
          corpus1: [
            {
              id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
              state: 'open',
              name: 'Shin Seiki Evangelion',
              description: 'Become legend, young boy!',
              sets: ['unit-00', 'unit-01', 'unit-02'],
              corpus_id: 'corpus1'
            },
            {
              id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
              state: 'open',
              name: 'Mobile Suit Gundam Wing',
              description: 'another dataset',
              sets: ['training', 'test', 'validation'],
              corpus_id: 'corpus1'
            }
          ]
        }
        const responseData = {
          id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
          state: 'open',
          name: 'Pretty Guardian Sailor Moon',
          description: 'another dataset',
          sets: ['training', 'test', 'validation']
        }
        mock.onPatch('/datasets/bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb/', { name: 'Pretty Guardian Sailor Moon' }).reply(200, { ...responseData })

        await store.dispatch('corpora/updateCorpusDataset', {
          corpusId: 'corpus1',
          datasetId: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
          name: 'Pretty Guardian Sailor Moon'
        })

        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/updateCorpusDataset',
            payload: {
              corpusId: 'corpus1',
              datasetId: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
              name: 'Pretty Guardian Sailor Moon'
            }
          },
          { mutation: 'corpora/updateCorpusDatasets', payload: { corpusId: 'corpus1', data: responseData } }
        ])

        assert.deepStrictEqual(store.state.corpora.corpusDatasets, {
          corpus1: [
            {
              id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
              state: 'open',
              name: 'Shin Seiki Evangelion',
              description: 'Become legend, young boy!',
              sets: ['unit-00', 'unit-01', 'unit-02'],
              corpus_id: 'corpus1'
            },
            {
              id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
              state: 'open',
              name: 'Pretty Guardian Sailor Moon',
              description: 'another dataset',
              sets: ['training', 'test', 'validation'],
              corpus_id: 'corpus1'
            }
          ]
        })
      })

      it('handles errors', async () => {
        mock.onPatch('/datasets/bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb/', { name: 'Pretty Guardian Sailor Moon' }).reply(400)
        await assertRejects(async () => store.dispatch('corpora/updateCorpusDataset', {
          corpusId: 'corpus1',
          datasetId: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
          name: 'Pretty Guardian Sailor Moon'
        }))

        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/updateCorpusDataset',
            payload: {
              corpusId: 'corpus1',
              datasetId: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
              name: 'Pretty Guardian Sailor Moon'
            }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 400'
            }
          }
        ])
      })
    })

    describe('deleteCorpusDataset', () => {
      it('deletes an existing dataset from a corpus', async () => {
        store.state.corpora.corpusDatasets = {
          corpus1: [
            {
              id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
              state: 'open',
              name: 'Shin Seiki Evangelion',
              description: 'Become legend, young boy!',
              sets: ['unit-00', 'unit-01', 'unit-02'],
              corpus_id: 'corpus1'
            },
            {
              id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
              state: 'open',
              name: 'Mobile Suit Gundam Wing',
              description: 'another dataset',
              sets: ['training', 'test', 'validation'],
              corpus_id: 'corpus1'
            }
          ]
        }
        mock.onDelete('/datasets/bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb/').reply(204)

        await store.dispatch('corpora/deleteCorpusDataset', {
          corpusId: 'corpus1',
          datasetId: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb'
        })

        assert.deepStrictEqual(store.history, [
          { action: 'corpora/deleteCorpusDataset', payload: { corpusId: 'corpus1', datasetId: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb' } },
          { mutation: 'corpora/removeCorpusDataset', payload: { corpusId: 'corpus1', datasetId: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb' } }
        ])

        assert.deepStrictEqual(store.state.corpora.corpusDatasets, {
          corpus1: [
            {
              id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
              state: 'open',
              name: 'Shin Seiki Evangelion',
              description: 'Become legend, young boy!',
              sets: ['unit-00', 'unit-01', 'unit-02'],
              corpus_id: 'corpus1'
            }
          ]
        })
      })

      it('handles errors', async () => {
        mock.onDelete('/datasets/bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb/').reply(400)
        await assertRejects(async () => store.dispatch('corpora/deleteCorpusDataset', {
          corpusId: 'corpus1',
          datasetId: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb'
        }))

        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/deleteCorpusDataset',
            payload: {
              corpusId: 'corpus1',
              datasetId: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb'
            }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 400'
            }
          }
        ])
      })
    })

    describe('addDatasetElementsSelection', () => {
      it('adds elements from a selection to a dataset', async () => {
        mock.onPost('/corpus/corpusid/datasets/selection/', { dataset_id: 'datasetid', set: 'test' }).reply(204)

        const params = { dataset_id: 'datasetid', set: 'test' }
        await store.dispatch('corpora/addDatasetElementsSelection', { corpusId: 'corpusid', params })

        assert.deepStrictEqual(store.history, [
          { action: 'corpora/addDatasetElementsSelection', payload: { corpusId: 'corpusid', params } },
          { mutation: 'notifications/notify', payload: { type: 'success', text: 'Elements have been added to the dataset.' } }
        ])
      })

      it('throw errors', async () => {
        mock.onPost('/corpus/corpusid/datasets/selection/', { dataset_id: 'datasetid', set: 'test' }).reply(400)

        const params = { dataset_id: 'datasetid', set: 'test' }
        await assertRejects(
          async () => await store.dispatch('corpora/addDatasetElementsSelection', { corpusId: 'corpusid', params })
        )

        assert.deepStrictEqual(store.history, [
          { action: 'corpora/addDatasetElementsSelection', payload: { corpusId: 'corpusid', params } },
          { mutation: 'notifications/notify', payload: { type: 'error', text: 'Request failed with status code 400' } }
        ])
      })
    })

    describe('listExports', () => {
      it('lists exports in a corpus', async () => {
        mock.onGet('/corpus/corpusid/export/').reply(200, { count: 1, results: [exportSample] })
        await store.dispatch('corpora/listExports', { corpusId: 'corpusid' })
        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/listExports',
            payload: { corpusId: 'corpusid' }
          },
          {
            mutation: 'corpora/setExports',
            payload: { count: 1, results: [exportSample] }
          }
        ])
        assert.deepStrictEqual(store.state.corpora.exports, {
          count: 1,
          results: [exportSample]
        })
      })

      it('handles errors', async () => {
        mock.onGet('/corpus/corpusid/export/').reply(451, { detail: 'please allow cookies' })
        await store.dispatch('corpora/listExports', { corpusId: 'corpusid' })
        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/listExports',
            payload: { corpusId: 'corpusid' }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'please allow cookies'
            }
          }
        ])
      })
    })

    describe('startExport', () => {
      it('starts an export', async () => {
        mock.onPost('/corpus/corpusid/export/').reply(201, exportSample)
        mock.onGet('/jobs/').reply(200, jobsSample)

        await store.dispatch('corpora/startExport', 'corpusid')
        await store.actionsCompleted()

        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/startExport',
            payload: 'corpusid'
          }
        ])
        assert.deepStrictEqual(jobsStore.jobs, {
          jobid: jobsSample[0]
        })
      })

      it('handles errors', async () => {
        mock.onPost('/corpus/corpusid/export/').reply(402, { detail: 'not enough bitcoins' })

        await store.dispatch('corpora/startExport', 'corpusid')
        await store.actionsCompleted()

        assert.deepStrictEqual(store.history, [
          {
            action: 'corpora/startExport',
            payload: 'corpusid'
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'not enough bitcoins'
            }
          }
        ])
      })
    })
  })
})
