import { assert } from 'chai'
import axios from 'axios'
import { initialState, mutations } from '@/store/entity.js'
import store from './index.spec.js'
import {
  entitySample,
  entitiesSample,
  linksSample,
  elementSample,
  entitiesInTranscriptionSample
} from '../samples.js'
import { FakeAxios, assertRejects } from '../testhelpers.js'

describe('entity', () => {
  describe('mutations', () => {
    it('set', () => {
      const state = initialState()
      mutations.set(state, entitySample)
      assert.deepStrictEqual(state.entity, entitySample)
    })

    it('setEntities', () => {
      const state = initialState()
      mutations.setEntities(state, entitiesSample)
      assert.deepStrictEqual(state.entities, entitiesSample)
    })

    it('setLinks', () => {
      const state = initialState()
      mutations.setLinks(state, linksSample)
      assert.deepStrictEqual(state.links, linksSample)
    })

    it('setElements', () => {
      const state = initialState()
      mutations.setElements(state, elementSample)
      assert.deepStrictEqual(state.elements, elementSample)
    })

    it('setInTranscription', () => {
      const state = initialState()
      mutations.setInTranscription(state, {
        id: 'transcriptionid',
        ...entitiesInTranscriptionSample
      })
      assert.deepStrictEqual(state.inTranscription, {
        transcriptionid: {
          results: entitiesInTranscriptionSample.results,
          count: 1,
          loaded: 1
        }
      })
    })

    it('resetInTranscription', () => {
      const state = {
        inTranscription: {
          transcriptionid: {
            results: entitiesInTranscriptionSample.results,
            count: 1,
            loaded: 1
          }
        }
      }
      mutations.resetInTranscription(state, 'transcriptionid')
      assert.deepStrictEqual(state.inTranscription.transcriptionid, {
        results: null,
        loaded: null,
        count: null
      })
    })

    it('reset', () => {
      const state = {
        entity: entitySample,
        entities: entitiesSample,
        links: linksSample,
        elements: elementSample,
        inTranscription: {
          transcriptionid: {
            results: entitiesInTranscriptionSample,
            count: 1,
            loaded: 1
          }
        }
      }
      mutations.reset(state)
      assert.deepStrictEqual(state, initialState())
    })
  })

  describe('actions', () => {
    let mock

    before('Setting up Axios mock', () => {
      mock = new FakeAxios(axios)
    })

    afterEach(() => {
      // Remove any handlers, but leave mocking in place
      mock.reset()
      store.reset()
    })

    after('Removing Axios mock', () => {
      // Remove mocking entirely
      mock.restore()
    })

    describe('get', () => {
      it('retrieves an entity', async () => {
        mock.onGet('/entity/entityid/').reply(200, entitySample)

        await store.dispatch('entity/get', { id: 'entityid' })

        assert.deepStrictEqual(store.history, [
          {
            action: 'entity/get',
            payload: { id: 'entityid' }
          },
          {
            mutation: 'entity/set',
            payload: entitySample
          }
        ])
      })

      it('handles errors', async () => {
        mock.onGet('/entity/entityid/').reply(500)

        await assertRejects(() => store.dispatch('entity/get', { id: 'entityid' }))

        assert.deepStrictEqual(store.history, [
          {
            action: 'entity/get',
            payload: { id: 'entityid' }
          },
          {
            mutation: 'entity/set',
            payload: null
          }
        ])
      })
    })

    describe('list', () => {
      it('lists entities', async () => {
        mock.onGet('/corpus/corpusid/entities/').reply(200, entitiesSample)

        await store.dispatch('entity/list', { id: 'corpusid' })

        assert.deepStrictEqual(store.history, [
          {
            action: 'entity/list',
            payload: { id: 'corpusid' }
          },
          {
            mutation: 'entity/setEntities',
            payload: entitiesSample
          }
        ])
      })

      it('handles errors', async () => {
        mock.onGet('/corpus/corpusid/entities/').reply(451, { detail: 'PC LOAD LETTER' })

        const error = await assertRejects(() => store.dispatch('entity/list', { id: 'corpusid' }))

        assert.deepStrictEqual(error.response.data, { detail: 'PC LOAD LETTER' })

        assert.deepStrictEqual(store.history, [
          {
            action: 'entity/list',
            payload: { id: 'corpusid' }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'PC LOAD LETTER'
            }
          }
        ])
      })
    })

    it('listLinks', async () => {
      mock.onGet('/element/elementid/links/').reply(200, linksSample)

      await store.dispatch('entity/listLinks', { id: 'elementid' })

      assert.deepStrictEqual(store.history, [
        {
          action: 'entity/listLinks',
          payload: { id: 'elementid' }
        },
        {
          mutation: 'entity/setLinks',
          payload: linksSample
        }
      ])
    })

    describe('listElements', () => {
      it('lists elements related to an entity', async () => {
        mock.onGet('/entity/entityid/elements/').reply(200, elementSample)

        await store.dispatch('entity/listElements', { id: 'entityid' })

        assert.deepStrictEqual(store.history, [
          {
            action: 'entity/listElements',
            payload: { id: 'entityid' }
          },
          {
            mutation: 'entity/setElements',
            payload: elementSample
          }
        ])
      })

      it('handles errors', async () => {
        mock.onGet('/entity/entityid/elements/').reply(500)

        await store.dispatch('entity/listElements', { id: 'entityid' })

        assert.deepStrictEqual(store.history, [
          {
            action: 'entity/listElements',
            payload: { id: 'entityid' }
          },
          {
            mutation: 'entity/setElements',
            payload: null
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 500'
            }
          }
        ])
      })
    })

    describe('listInTranscription', () => {
      it('lists entities in a transcription', async () => {
        mock.onGet('/transcription/transcriptionid/entities/').reply(200, entitiesInTranscriptionSample)

        await store.dispatch('entity/listInTranscription', { transcriptionId: 'transcriptionid' })

        assert.deepStrictEqual(store.history, [
          {
            action: 'entity/listInTranscription',
            payload: { transcriptionId: 'transcriptionid' }
          },
          {
            mutation: 'entity/resetInTranscription',
            payload: 'transcriptionid'
          },
          {
            mutation: 'entity/setInTranscription',
            payload: {
              id: 'transcriptionid',
              ...entitiesInTranscriptionSample
            }
          }
        ])
      })

      it('paginates through entities in a transcription', async () => {
        const firstPage = {
          ...entitiesInTranscriptionSample,
          next: '/transcription/transcriptionid/entities/?page=2'
        }
        mock.onGet('/transcription/transcriptionid/entities/', { params: { page: 1 } }).reply(200, firstPage)
        mock.onGet('/transcription/transcriptionid/entities/', { params: { page: 2 } }).reply(200, entitiesInTranscriptionSample)

        await store.dispatch('entity/listInTranscription', { transcriptionId: 'transcriptionid' })
        await store.actionsCompleted()

        assert.deepStrictEqual(store.history, [
          {
            action: 'entity/listInTranscription',
            payload: {
              transcriptionId: 'transcriptionid'
            }
          },
          {
            mutation: 'entity/resetInTranscription',
            payload: 'transcriptionid'
          },
          {
            mutation: 'entity/setInTranscription',
            payload: {
              id: 'transcriptionid',
              ...firstPage
            }
          },
          {
            action: 'entity/listInTranscription',
            payload: {
              transcriptionId: 'transcriptionid',
              page: 2
            }
          },
          {
            mutation: 'entity/setInTranscription',
            payload: {
              id: 'transcriptionid',
              ...entitiesInTranscriptionSample
            }
          }
        ])
      })

      it('handles errors', async () => {
        const firstPage = {
          ...entitiesInTranscriptionSample,
          next: '/transcription/transcriptionid/entities/?page=2'
        }
        mock.onGet('/transcription/transcriptionid/entities/', { params: { page: 1 } }).reply(200, firstPage)
        mock.onGet('/transcription/transcriptionid/entities/', { params: { page: 2 } }).reply(500)

        await store.dispatch('entity/listInTranscription', { transcriptionId: 'transcriptionid' })
        await store.actionsCompleted()

        assert.deepStrictEqual(store.history, [
          {
            action: 'entity/listInTranscription',
            payload: {
              transcriptionId: 'transcriptionid'
            }
          },
          {
            mutation: 'entity/resetInTranscription',
            payload: 'transcriptionid'
          },
          {
            mutation: 'entity/setInTranscription',
            payload: {
              id: 'transcriptionid',
              ...firstPage
            }
          },
          {
            action: 'entity/listInTranscription',
            payload: {
              transcriptionId: 'transcriptionid',
              page: 2
            }
          },
          {
            mutation: 'entity/resetInTranscription',
            payload: 'transcriptionid'
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 500'
            }
          }
        ])
      })
    })
  })
})
