import { assert } from 'chai'
import axios from 'axios'
import { initialState, mutations } from '@/store/repos.js'
import store from './index.spec.js'
import { repoSample, reposSample, availableReposSample, processSample } from '../samples.js'
import { assertRejects, FakeAxios } from '../testhelpers.js'

describe('repos', () => {
  describe('mutations', () => {
    it('setRepos', () => {
      const state = initialState()
      mutations.setRepos(state, [
        repoSample,
        { id: 'second_repo', name: 'repo2' }
      ])
      assert.deepStrictEqual(state.repositories, {
        repoid: repoSample,
        second_repo: { id: 'second_repo', name: 'repo2' }
      })
    })

    it('removeRepo', () => {
      const state = {
        repositories: { repoid: repoSample }
      }
      mutations.removeRepo(state, 'repoid')
      assert.deepStrictEqual(state.repositories, {})
    })

    it('setAvailable', () => {
      const state = initialState()
      mutations.setAvailable(state, availableReposSample)
      assert.deepStrictEqual(state.available, availableReposSample)
    })

    it('reset', () => {
      const state = { available: availableReposSample }
      mutations.reset(state)
      assert.deepStrictEqual(state, initialState())
    })
  })

  describe('actions', () => {
    let mock

    before('Setting up Axios mock', () => {
      mock = new FakeAxios(axios)
    })

    afterEach(() => {
      // Remove any handlers, but leave mocking in place
      mock.reset()
      store.reset()
    })

    after('Removing Axios mock', () => {
      // Remove mocking entirely
      mock.restore()
    })

    describe('list', () => {
      it('lists existing repositories', async () => {
        const sample = {
          ...reposSample,
          count: 2,
          results: [
            ...reposSample.results,
            {
              id: 'other_repo',
              workers: [
                {
                  id: 'worker 2'
                }, {
                  id: 'worker 3'
                }
              ]
            }
          ]
        }
        mock.onGet('/process/repos/').reply(200, sample)

        const page = await store.dispatch('repos/list', {})

        assert.deepStrictEqual(store.history, [
          { action: 'repos/list', payload: {} },
          { mutation: 'repos/setRepos', payload: sample.results },
          { mutation: 'process/setWorkers', payload: [{ id: 'worker 1' }, { id: 'worker 2' }, { id: 'worker 3' }] }
        ])
        assert.deepStrictEqual(page, sample)
        assert.deepStrictEqual(Object.keys(store.state.process.workers), ['worker 1', 'worker 2', 'worker 3'])
      })
    })

    describe('listAvailable', () => {
      it('lists available repositories', async () => {
        mock.onGet('/process/repos/search/credid/').reply(200, availableReposSample)

        await store.dispatch('repos/listAvailable', { credential: 'credid' })

        assert.deepStrictEqual(store.history, [
          {
            action: 'repos/listAvailable',
            payload: { credential: 'credid' }
          },
          {
            mutation: 'repos/setAvailable',
            payload: availableReposSample
          }
        ])

        assert.deepStrictEqual(store.state.repos, {
          repositories: {},
          available: availableReposSample
        })
      })

      it('searches with search terms', async () => {
        mock.onGet('/process/repos/search/credid/', { params: { search: 'blah' } }).reply(200, availableReposSample)

        await store.dispatch('repos/listAvailable', { credential: 'credid', search: 'blah' })

        assert.deepStrictEqual(store.history, [
          {
            action: 'repos/listAvailable',
            payload: {
              credential: 'credid',
              search: 'blah'
            }
          },
          {
            mutation: 'repos/setAvailable',
            payload: availableReposSample
          }
        ])

        assert.deepStrictEqual(store.state.repos, {
          repositories: {},
          available: availableReposSample
        })
      })
    })

    describe('create', () => {
      it('creates a new repo and a process', async () => {
        mock.onPost('/process/repos/search/credid/').reply(201, processSample)

        await store.dispatch('repos/create', { credential: 'credid', data: { id: 123456 } })

        assert.deepStrictEqual(store.history, [
          {
            action: 'repos/create',
            payload: {
              credential: 'credid',
              data: { id: 123456 }
            }
          },
          {
            mutation: 'process/setProcess',
            payload: processSample
          }
        ])
      })

      it('handles errors', async () => {
        mock.onPost('/process/repos/search/credid/').reply(400, { detail: 'oh no' })

        await assertRejects(
          async () => store.dispatch('repos/create', { credential: 'credid', data: { id: 123456 } })
        )

        assert.deepStrictEqual(store.history, [
          {
            action: 'repos/create',
            payload: {
              credential: 'credid',
              data: { id: 123456 }
            }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'oh no'
            }
          }
        ])
      })
    })

    describe('delete', () => {
      it('deletes a repo', async () => {
        mock.onDelete('/process/repos/repoid/').reply(204)

        store.state.repositories = { repoid: {}, secondId: {} }
        await store.dispatch('repos/delete', 'repoid')

        assert.deepStrictEqual(store.history, [
          {
            action: 'repos/delete',
            payload: 'repoid'
          },
          {
            mutation: 'repos/removeRepo',
            payload: 'repoid'
          }
        ])
      })
    })
  })
})
